var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _InnerModelDB_models;
import * as Comlink from "comlink";
import { assert, signalInvalidType } from "@canvas-js/utils";
import { ModelAPI } from "./ModelAPI.js";
export class InnerModelDB {
    constructor(db, config, log) {
        Object.defineProperty(this, "db", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        _InnerModelDB_models.set(this, {});
        Object.defineProperty(this, "log", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.db = db;
        for (const model of Object.values(config.models)) {
            __classPrivateFieldGet(this, _InnerModelDB_models, "f")[model.name] = new ModelAPI(this.db, model);
        }
        this.log = log;
    }
    close() {
        this.db.close();
    }
    apply(effects) {
        this.db.transaction(() => {
            for (const effect of effects) {
                const model = __classPrivateFieldGet(this, _InnerModelDB_models, "f")[effect.model];
                assert(model !== undefined, `model ${effect.model} not found`);
                if (effect.operation === "set") {
                    __classPrivateFieldGet(this, _InnerModelDB_models, "f")[effect.model].set(effect.value);
                }
                else if (effect.operation === "delete") {
                    __classPrivateFieldGet(this, _InnerModelDB_models, "f")[effect.model].delete(effect.key);
                }
                else {
                    signalInvalidType(effect);
                }
            }
        });
    }
    get(modelName, key) {
        const api = __classPrivateFieldGet(this, _InnerModelDB_models, "f")[modelName];
        assert(api !== undefined, `model ${modelName} not found`);
        return api.get(key);
    }
    getMany(modelName, keys) {
        const api = __classPrivateFieldGet(this, _InnerModelDB_models, "f")[modelName];
        assert(api !== undefined, `model ${modelName} not found`);
        return api.getMany(keys);
    }
    iterate(modelName, query = {}) {
        const api = __classPrivateFieldGet(this, _InnerModelDB_models, "f")[modelName];
        assert(api !== undefined, `model ${modelName} not found`);
        return Comlink.proxy(api.iterate(query));
    }
    count(modelName, where) {
        const api = __classPrivateFieldGet(this, _InnerModelDB_models, "f")[modelName];
        assert(api !== undefined, `model ${modelName} not found`);
        return api.count(where);
    }
    query(modelName, query = {}) {
        const api = __classPrivateFieldGet(this, _InnerModelDB_models, "f")[modelName];
        assert(api !== undefined, `model ${modelName} not found`);
        return api.query(query);
    }
    clear(modelName) {
        const api = __classPrivateFieldGet(this, _InnerModelDB_models, "f")[modelName];
        assert(api !== undefined, `model ${modelName} not found`);
        return api.clear();
    }
}
_InnerModelDB_models = new WeakMap();
