var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var _ModelAPI_table, _ModelAPI_params, _ModelAPI_properties, _ModelAPI_insert, _ModelAPI_update, _ModelAPI_delete, _ModelAPI_clear, _ModelAPI_selectAll, _ModelAPI_select, _ModelAPI_count, _ModelAPI_relations, _ModelAPI_primaryKeyName, _ModelAPI_primaryKeyParam, _RelationAPI_select, _RelationAPI_insert, _RelationAPI_delete;
import { assert, signalInvalidType, mapValues } from "@canvas-js/utils";
import { isNotExpression, isLiteralExpression, isRangeExpression, isPrimitiveValue, validateModelValue, } from "@canvas-js/modeldb";
import { zip } from "@canvas-js/utils";
import { decodePrimaryKeyValue, decodePrimitiveValue, decodeRecord, decodeReferenceValue, encodeRecordParams, } from "./encoding.js";
import { Method, Query } from "./utils.js";
const primitiveColumnTypes = {
    integer: "INTEGER",
    float: "FLOAT",
    number: "NUMERIC",
    string: "TEXT",
    bytes: "BLOB",
    boolean: "INTEGER",
    json: "TEXT",
};
function getPropertyColumnType(property) {
    if (property.kind === "primary") {
        return "TEXT PRIMARY KEY NOT NULL";
    }
    else if (property.kind === "primitive") {
        const type = primitiveColumnTypes[property.type];
        return property.nullable ? type : `${type} NOT NULL`;
    }
    else if (property.kind === "reference") {
        return property.nullable ? "TEXT" : "TEXT NOT NULL";
    }
    else if (property.kind === "relation") {
        throw new Error("internal error - relation properties don't map to columns");
    }
    else {
        signalInvalidType(property);
    }
}
const getPropertyColumn = (property) => `'${property.name}' ${getPropertyColumnType(property)}`;
export class ModelAPI {
    constructor(db, model) {
        Object.defineProperty(this, "db", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: db
        });
        Object.defineProperty(this, "model", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: model
        });
        _ModelAPI_table.set(this, this.model.name);
        _ModelAPI_params.set(this, {});
        _ModelAPI_properties.set(this, Object.fromEntries(this.model.properties.map((property) => [property.name, property]))
        // Methods
        );
        // Methods
        _ModelAPI_insert.set(this, void 0);
        _ModelAPI_update.set(this, void 0);
        _ModelAPI_delete.set(this, void 0);
        _ModelAPI_clear.set(this, void 0);
        // Queries
        _ModelAPI_selectAll.set(this, void 0);
        _ModelAPI_select.set(this, void 0);
        _ModelAPI_count.set(this, void 0);
        _ModelAPI_relations.set(this, {});
        _ModelAPI_primaryKeyName.set(this, void 0);
        _ModelAPI_primaryKeyParam.set(this, void 0);
        Object.defineProperty(this, "columnNames", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        const columns = [];
        this.columnNames = []; // quoted column names for non-relation properties
        const columnParams = []; // query params for non-relation properties
        let primaryKeyIndex = null;
        let primaryKey = null;
        for (const [i, property] of model.properties.entries()) {
            if (property.kind === "primary" || property.kind === "primitive" || property.kind === "reference") {
                columns.push(getPropertyColumn(property));
                this.columnNames.push(`"${property.name}"`);
                columnParams.push(`:p${i}`);
                __classPrivateFieldGet(this, _ModelAPI_params, "f")[property.name] = `p${i}`;
                if (property.kind === "primary") {
                    primaryKeyIndex = i;
                    primaryKey = property;
                }
            }
            else if (property.kind === "relation") {
                __classPrivateFieldGet(this, _ModelAPI_relations, "f")[property.name] = new RelationAPI(db, {
                    source: model.name,
                    property: property.name,
                    target: property.target,
                    indexed: false,
                });
            }
            else {
                signalInvalidType(property);
            }
        }
        assert(primaryKey !== null, "expected primaryKey !== null");
        assert(primaryKeyIndex !== null, "expected primaryKeyIndex !== null");
        // this.#primaryKeyName = columnNames[primaryKeyIndex]
        __classPrivateFieldSet(this, _ModelAPI_primaryKeyName, primaryKey.name, "f");
        __classPrivateFieldSet(this, _ModelAPI_primaryKeyParam, `p${primaryKeyIndex}`, "f");
        // Create record table
        db.exec(`CREATE TABLE IF NOT EXISTS "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}" (${columns.join(", ")})`);
        // Create indexes
        for (const index of model.indexes) {
            const indexName = `${model.name}/${index.join("/")}`;
            const indexColumns = index.map((name) => `'${name}'`);
            db.exec(`CREATE INDEX IF NOT EXISTS "${indexName}" ON "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}" (${indexColumns.join(", ")})`);
        }
        // Prepare methods
        const insertNames = this.columnNames.join(", ");
        const insertParams = columnParams.join(", ");
        __classPrivateFieldSet(this, _ModelAPI_insert, new Method(db, `INSERT OR IGNORE INTO "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}" (${insertNames}) VALUES (${insertParams})`), "f");
        const where = `WHERE "${__classPrivateFieldGet(this, _ModelAPI_primaryKeyName, "f")}" = :${__classPrivateFieldGet(this, _ModelAPI_primaryKeyParam, "f")}`;
        const updateEntries = Array.from(zip(this.columnNames, columnParams)).map(([name, param]) => `${name} = ${param}`);
        __classPrivateFieldSet(this, _ModelAPI_update, new Method(db, `UPDATE "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}" SET ${updateEntries.join(", ")} ${where}`), "f");
        __classPrivateFieldSet(this, _ModelAPI_delete, new Method(db, `DELETE FROM "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}" ${where}`), "f");
        __classPrivateFieldSet(this, _ModelAPI_clear, new Method(db, `DELETE FROM "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}"`), "f");
        // Prepare queries
        __classPrivateFieldSet(this, _ModelAPI_count, new Query(this.db, `SELECT COUNT(*) AS count FROM "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}"`), "f");
        __classPrivateFieldSet(this, _ModelAPI_select, new Query(this.db, `SELECT ${this.columnNames.join(", ")} FROM "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}" ${where}`), "f");
        __classPrivateFieldSet(this, _ModelAPI_selectAll, new Query(this.db, `SELECT ${this.columnNames.join(", ")} FROM "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}"`), "f");
    }
    get(key) {
        return this.getMany([key])[0];
    }
    getMany(keys) {
        if (keys.length === 0) {
            return [];
        }
        const params = {};
        const whereParts = [];
        for (const [i, key] of keys.entries()) {
            whereParts.push(`"${__classPrivateFieldGet(this, _ModelAPI_primaryKeyName, "f")}" = :p${i}`);
            params[`p${i}`] = key;
        }
        const queryString = `SELECT ${this.columnNames.join(", ")} FROM "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}" WHERE ${whereParts.join(" OR ")}`;
        const query = new Query(this.db, queryString);
        const rowsByKey = {};
        for (const row of query.all(params)) {
            const rowKey = row[__classPrivateFieldGet(this, _ModelAPI_primaryKeyName, "f")];
            assert(typeof rowKey === "string", 'expected typeof primaryKey === "string"');
            rowsByKey[rowKey] = {
                ...decodeRecord(this.model, row),
                ...mapValues(__classPrivateFieldGet(this, _ModelAPI_relations, "f"), (api) => api.get(rowKey)),
            };
        }
        return keys.map((key) => rowsByKey[key] ?? null);
    }
    set(value) {
        validateModelValue(this.model, value);
        const key = value[__classPrivateFieldGet(this, _ModelAPI_primaryKeyName, "f")];
        assert(typeof key === "string", 'expected typeof primaryKey === "string"');
        const encodedParams = encodeRecordParams(this.model, value, __classPrivateFieldGet(this, _ModelAPI_params, "f"));
        const existingRecord = __classPrivateFieldGet(this, _ModelAPI_select, "f").get({ [__classPrivateFieldGet(this, _ModelAPI_primaryKeyParam, "f")]: key });
        if (existingRecord === null) {
            __classPrivateFieldGet(this, _ModelAPI_insert, "f").run(encodedParams);
        }
        else {
            __classPrivateFieldGet(this, _ModelAPI_update, "f").run(encodedParams);
        }
        for (const [name, relation] of Object.entries(__classPrivateFieldGet(this, _ModelAPI_relations, "f"))) {
            if (existingRecord !== null) {
                relation.delete(key);
            }
            relation.add(key, value[name]);
        }
    }
    delete(key) {
        const existingRecord = __classPrivateFieldGet(this, _ModelAPI_select, "f").get({ [__classPrivateFieldGet(this, _ModelAPI_primaryKeyParam, "f")]: key });
        if (existingRecord === null) {
            return;
        }
        __classPrivateFieldGet(this, _ModelAPI_delete, "f").run({ [__classPrivateFieldGet(this, _ModelAPI_primaryKeyParam, "f")]: key });
        for (const relation of Object.values(__classPrivateFieldGet(this, _ModelAPI_relations, "f"))) {
            relation.delete(key);
        }
    }
    clear() {
        const existingRecords = __classPrivateFieldGet(this, _ModelAPI_select, "f").all({});
        __classPrivateFieldGet(this, _ModelAPI_clear, "f").run({});
        for (const record of existingRecords) {
            const key = record[__classPrivateFieldGet(this, _ModelAPI_primaryKeyParam, "f")];
            for (const relation of Object.values(__classPrivateFieldGet(this, _ModelAPI_relations, "f"))) {
                if (!key || typeof key !== "string")
                    continue;
                relation.delete(key);
            }
        }
    }
    count(where) {
        const sql = [];
        // SELECT
        sql.push(`SELECT COUNT(*) AS count FROM "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}"`);
        // WHERE
        const [whereExpression, params] = this.getWhereExpression(where);
        if (whereExpression) {
            sql.push(`WHERE ${whereExpression}`);
        }
        const paramsWithColons = {};
        for (const key of Object.keys(params)) {
            paramsWithColons[`:${key}`] = params[key];
        }
        const results = this.db.selectObjects(sql.join(" "), paramsWithColons);
        const countResult = results[0].count;
        if (typeof countResult === "number") {
            return countResult;
        }
        else {
            throw new Error("internal error");
        }
    }
    async *iterate(query = {}) {
        // TODO: implement iterate (needs special handling over comlink)
        if (Object.keys(query).length > 0) {
            throw new Error("not implemented");
        }
        for (const record of __classPrivateFieldGet(this, _ModelAPI_selectAll, "f").iterate({})) {
            const key = record[__classPrivateFieldGet(this, _ModelAPI_primaryKeyName, "f")];
            assert(typeof key === "string", 'expected typeof key === "string"');
            const value = {
                ...decodeRecord(this.model, record),
                ...mapValues(__classPrivateFieldGet(this, _ModelAPI_relations, "f"), (api) => api.get(key)),
            };
            yield value;
        }
    }
    query(query) {
        const [sql, relations, params] = this.parseQuery(query);
        const paramsWithColons = {};
        for (const key of Object.keys(params)) {
            paramsWithColons[`:${key}`] = params[key];
        }
        const results = this.db.selectObjects(sql, paramsWithColons);
        return results.map((record) => {
            const key = record[__classPrivateFieldGet(this, _ModelAPI_primaryKeyName, "f")];
            assert(typeof key === "string", 'expected typeof primaryKey === "string"');
            const value = {};
            for (const [propertyName, propertyValue] of Object.entries(record)) {
                const property = __classPrivateFieldGet(this, _ModelAPI_properties, "f")[propertyName];
                if (property.kind === "primary") {
                    value[propertyName] = decodePrimaryKeyValue(this.model.name, property, propertyValue);
                }
                else if (property.kind === "primitive") {
                    value[propertyName] = decodePrimitiveValue(this.model.name, property, propertyValue);
                }
                else if (property.kind === "reference") {
                    value[propertyName] = decodeReferenceValue(this.model.name, property, propertyValue);
                }
                else if (property.kind === "relation") {
                    throw new Error("internal error");
                }
                else {
                    signalInvalidType(property);
                }
            }
            for (const relation of relations) {
                value[relation.property] = __classPrivateFieldGet(this, _ModelAPI_relations, "f")[relation.property].get(key);
            }
            return value;
        });
    }
    parseQuery(query) {
        // See https://www.sqlite.org/lang_select.html for railroad diagram
        const sql = [];
        // SELECT
        const [select, relations] = this.getSelectExpression(query.select);
        sql.push(`SELECT ${select} FROM "${__classPrivateFieldGet(this, _ModelAPI_table, "f")}"`);
        // WHERE
        const [where, params] = this.getWhereExpression(query.where);
        if (where !== null) {
            sql.push(`WHERE ${where}`);
        }
        // ORDER BY
        if (query.orderBy !== undefined) {
            const orders = Object.entries(query.orderBy);
            assert(orders.length === 1, "cannot order by multiple properties at once");
            const [[name, direction]] = orders;
            const property = __classPrivateFieldGet(this, _ModelAPI_properties, "f")[name];
            assert(property !== undefined, "property not found");
            assert(property.kind === "primary" || property.kind === "primitive" || property.kind === "reference", "cannot order by relation properties");
            if (direction === "asc") {
                sql.push(`ORDER BY "${name}" ASC`);
            }
            else if (direction === "desc") {
                sql.push(`ORDER BY "${name}" DESC`);
            }
            else {
                throw new Error("invalid orderBy direction");
            }
        }
        // LIMIT
        if (typeof query.limit === "number") {
            sql.push(`LIMIT :limit`);
            params.limit = query.limit;
        }
        // OFFSET
        if (typeof query.offset === "number") {
            sql.push(`LIMIT :offset`);
            params.limit = query.offset;
        }
        // JOIN (not supported)
        if (query.include) {
            throw new Error("cannot use 'include' in queries outside the browser/idb");
        }
        return [sql.join(" "), relations, params];
    }
    getSelectExpression(select = mapValues(__classPrivateFieldGet(this, _ModelAPI_properties, "f"), () => true)) {
        const relations = [];
        const columns = [];
        for (const [name, value] of Object.entries(select)) {
            if (value === false) {
                continue;
            }
            const property = __classPrivateFieldGet(this, _ModelAPI_properties, "f")[name];
            assert(property !== undefined, "property not found");
            if (property.kind === "primary" || property.kind === "primitive" || property.kind === "reference") {
                columns.push(`"${name}"`);
            }
            else if (property.kind === "relation") {
                relations.push({
                    source: this.model.name,
                    property: name,
                    target: property.target,
                    indexed: this.model.indexes.some((index) => index.length === 1 && index[0] === name),
                });
            }
            else {
                signalInvalidType(property);
            }
        }
        assert(columns.length > 0, "cannot query an empty select expression");
        assert(columns.includes(`"${__classPrivateFieldGet(this, _ModelAPI_primaryKeyName, "f")}"`), "select expression must include the primary key");
        return [columns.join(", "), relations];
    }
    getWhereExpression(where = {}) {
        const params = {};
        const filters = Object.entries(where).flatMap(([name, expression], i) => {
            const property = __classPrivateFieldGet(this, _ModelAPI_properties, "f")[name];
            assert(property !== undefined, "property not found");
            if (expression === undefined) {
                return [];
            }
            if (property.kind === "primary") {
                if (isLiteralExpression(expression)) {
                    if (typeof expression !== "string") {
                        throw new TypeError("invalid primary key value (expected string)");
                    }
                    const p = `p${i}`;
                    params[p] = expression;
                    return [`"${name}" = :${p}`];
                }
                else if (isNotExpression(expression)) {
                    const { neq: value } = expression;
                    if (typeof value !== "string") {
                        throw new TypeError("invalid primary key value (expected string)");
                    }
                    const p = `p${i}`;
                    params[p] = value;
                    return [`"${name}" != :${p}`];
                }
                else if (isRangeExpression(expression)) {
                    const keys = Object.keys(expression);
                    return keys
                        .filter((key) => expression[key] !== undefined)
                        .flatMap((key, j) => {
                        const value = expression[key];
                        if (typeof value !== "string") {
                            throw new TypeError("invalid primary key value (expected string)");
                        }
                        const p = `p${i}q${j}`;
                        params[p] = value;
                        switch (key) {
                            case "gt":
                                return [`"${name}" > :${p}`];
                            case "gte":
                                return [`"${name}" >= :${p}`];
                            case "lt":
                                return [`"${name}" < :${p}`];
                            case "lte":
                                return [`"${name}" <= :${p}`];
                        }
                    });
                }
                else {
                    signalInvalidType(expression);
                }
            }
            else if (property.kind === "primitive") {
                if (property.type === "json") {
                    throw new Error("json properties are not supported in where clauses");
                }
                if (isLiteralExpression(expression)) {
                    if (expression === null) {
                        return [`"${name}" ISNULL`];
                    }
                    else if (Array.isArray(expression)) {
                        throw new Error("invalid primitive value (expected null | number | string | Uint8Array)");
                    }
                    else {
                        assert(isPrimitiveValue(expression));
                        const p = `p${i}`;
                        params[p] = expression;
                        return [`"${name}" = :${p}`];
                    }
                }
                else if (isNotExpression(expression)) {
                    const { neq: value } = expression;
                    if (value === undefined) {
                        return [];
                    }
                    else if (value === null) {
                        return [`"${name}" NOTNULL`];
                    }
                    else if (Array.isArray(value)) {
                        throw new Error("invalid primitive value (expected null | number | string | Uint8Array)");
                    }
                    else {
                        assert(isPrimitiveValue(value));
                        const p = `p${i}`;
                        params[p] = value;
                        if (property.nullable) {
                            return [`("${name}" ISNULL OR "${name}" != :${p})`];
                        }
                        else {
                            return [`"${name}" != :${p}`];
                        }
                    }
                }
                else if (isRangeExpression(expression)) {
                    const keys = Object.keys(expression);
                    return keys
                        .filter((key) => expression[key] !== undefined)
                        .flatMap((key, j) => {
                        const value = expression[key];
                        if (value === null) {
                            switch (key) {
                                case "gt":
                                    return [`"${name}" NOTNULL`];
                                case "gte":
                                    return [];
                                case "lt":
                                    return ["0 = 1"];
                                case "lte":
                                    return [];
                            }
                        }
                        const p = `p${i}q${j}`;
                        params[p] = value;
                        switch (key) {
                            case "gt":
                                return [`("${name}" NOTNULL) AND ("${name}" > :${p})`];
                            case "gte":
                                return [`("${name}" NOTNULL) AND ("${name}" >= :${p})`];
                            case "lt":
                                return [`("${name}" ISNULL) OR ("${name}" < :${p})`];
                            case "lte":
                                return [`("${name}" ISNULL) OR ("${name}" <= :${p})`];
                        }
                    });
                }
                else {
                    signalInvalidType(expression);
                }
            }
            else if (property.kind === "reference") {
                if (isLiteralExpression(expression)) {
                    const reference = expression;
                    if (reference === null) {
                        return [`"${name}" ISNULL`];
                    }
                    else if (typeof reference === "string") {
                        const p = `p${i}`;
                        params[p] = reference;
                        return [`"${name}" = :${p}`];
                    }
                    else {
                        throw new Error("invalid reference value (expected string | null)");
                    }
                }
                else if (isNotExpression(expression)) {
                    const reference = expression.neq;
                    if (reference === null) {
                        return [`"${name}" NOTNULL`];
                    }
                    else if (typeof reference === "string") {
                        const p = `p${i}`;
                        params[p] = reference;
                        return [`"${name}" != :${p}`];
                    }
                    else {
                        throw new Error("invalid reference value (expected string | null)");
                    }
                }
                else if (isRangeExpression(expression)) {
                    throw new Error("cannot use range expressions on reference values");
                }
                else {
                    signalInvalidType(expression);
                }
            }
            else if (property.kind === "relation") {
                const relationTable = __classPrivateFieldGet(this, _ModelAPI_relations, "f")[property.name].table;
                if (isLiteralExpression(expression)) {
                    const references = expression;
                    assert(Array.isArray(references), "invalid relation value (expected string[])");
                    const targets = [];
                    for (const [j, reference] of references.entries()) {
                        assert(typeof reference === "string", "invalid relation value (expected string[])");
                        const p = `p${i}q${j}`;
                        params[p] = reference;
                        targets.push(`"${__classPrivateFieldGet(this, _ModelAPI_primaryKeyName, "f")}" IN (SELECT _source FROM "${relationTable}" WHERE (_target = :${p}))`);
                    }
                    return targets.length > 0 ? [targets.join(" AND ")] : [];
                }
                else if (isNotExpression(expression)) {
                    const references = expression.neq;
                    assert(Array.isArray(references), "invalid relation value (expected string[])");
                    const targets = [];
                    for (const [j, reference] of references.entries()) {
                        assert(typeof reference === "string", "invalid relation value (expected string[])");
                        const p = `p${i}q${j}`;
                        params[p] = reference;
                        targets.push(`"${__classPrivateFieldGet(this, _ModelAPI_primaryKeyName, "f")}" NOT IN (SELECT _source FROM "${relationTable}" WHERE (_target = :${p}))`);
                    }
                    return targets.length > 0 ? [targets.join(" AND ")] : [];
                }
                else if (isRangeExpression(expression)) {
                    throw new Error("cannot use range expressions on relation values");
                }
                else {
                    signalInvalidType(expression);
                }
            }
            else {
                signalInvalidType(property);
            }
        });
        if (filters.length === 0) {
            return [null, {}];
        }
        else {
            return [`${filters.map((filter) => `(${filter})`).join(" AND ")}`, params];
        }
    }
}
_ModelAPI_table = new WeakMap(), _ModelAPI_params = new WeakMap(), _ModelAPI_properties = new WeakMap(), _ModelAPI_insert = new WeakMap(), _ModelAPI_update = new WeakMap(), _ModelAPI_delete = new WeakMap(), _ModelAPI_clear = new WeakMap(), _ModelAPI_selectAll = new WeakMap(), _ModelAPI_select = new WeakMap(), _ModelAPI_count = new WeakMap(), _ModelAPI_relations = new WeakMap(), _ModelAPI_primaryKeyName = new WeakMap(), _ModelAPI_primaryKeyParam = new WeakMap();
export class RelationAPI {
    constructor(db, relation) {
        Object.defineProperty(this, "db", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: db
        });
        Object.defineProperty(this, "relation", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: relation
        });
        Object.defineProperty(this, "table", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: `${this.relation.source}/${this.relation.property}`
        });
        Object.defineProperty(this, "sourceIndex", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: `${this.relation.source}/${this.relation.property}/source`
        });
        Object.defineProperty(this, "targetIndex", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: `${this.relation.source}/${this.relation.property}/target`
        });
        _RelationAPI_select.set(this, void 0);
        _RelationAPI_insert.set(this, void 0);
        _RelationAPI_delete.set(this, void 0);
        const columns = [`_source TEXT NOT NULL`, `_target TEXT NOT NULL`];
        db.exec(`CREATE TABLE IF NOT EXISTS "${this.table}" (${columns.join(", ")})`);
        db.exec(`CREATE INDEX IF NOT EXISTS "${this.sourceIndex}" ON "${this.table}" (_source)`);
        if (relation.indexed) {
            db.exec(`CREATE INDEX IF NOT EXISTS "${this.targetIndex}" ON "${this.table}" (_target)`);
        }
        // Prepare methods
        __classPrivateFieldSet(this, _RelationAPI_insert, new Method(this.db, `INSERT INTO "${this.table}" (_source, _target) VALUES (:_source, :_target)`), "f");
        __classPrivateFieldSet(this, _RelationAPI_delete, new Method(this.db, `DELETE FROM "${this.table}" WHERE _source = :_source`), "f");
        // Prepare queries
        __classPrivateFieldSet(this, _RelationAPI_select, new Query(this.db, `SELECT _target FROM "${this.table}" WHERE _source = :_source`), "f");
    }
    get(source) {
        const targets = __classPrivateFieldGet(this, _RelationAPI_select, "f").all({ _source: source });
        return targets.map(({ _target: target }) => target);
    }
    add(source, targets) {
        assert(Array.isArray(targets), "expected string[]");
        for (const target of targets) {
            assert(typeof target === "string", "expected string[]");
            __classPrivateFieldGet(this, _RelationAPI_insert, "f").run({ _source: source, _target: target });
        }
    }
    delete(source) {
        __classPrivateFieldGet(this, _RelationAPI_delete, "f").run({ _source: source });
    }
}
_RelationAPI_select = new WeakMap(), _RelationAPI_insert = new WeakMap(), _RelationAPI_delete = new WeakMap();
